import time

from sockjs.cyclone.utils import PriorityQueue


class SessionContainer(object):
    """ Session container object. """
    def __init__(self):
        self._items = dict()
        self._queue = PriorityQueue()

    def add(self, session):
        """ Add session to the container.

        @param session: Session object
        """
        self._items[session.session_id] = session

        if session.expiry is not None:
            self._queue.push(session)

    def get(self, session_id):
        """ Return session object or None if it is not available

        @param session_id: Session identifier
        """
        return self._items.get(session_id, None)

    def remove(self, session_id):
        """ Remove session object from the container

        @param session_id: Session identifier
        """
        session = self._items.get(session_id, None)

        if session is not None:
            session.promoted = -1
            session.on_delete(True)
            del self._items[session_id]
            return True

        return False

    def expire(self, current_time=None):
        """ Expire any old entries

        @param current_time: Optional time to be used to clean up queue (can be
                             used in unit tests)
        """
        if self._queue.is_empty():
            return

        if current_time is None:
            current_time = time.time()

        while not self._queue.is_empty():
            # Get top most item
            top = self._queue.peek()

            # Early exit if item was not promoted and its expiration time
            # is greater than now.
            if top.promoted is None and top.expiry_date > current_time:
                break

            # Pop item from the stack
            top = self._queue.pop()

            need_reschedule = (top.promoted is not None
                               and top.promoted > current_time)

            # Give chance to reschedule
            if not need_reschedule:
                top.promoted = None
                top.on_delete(False)

                need_reschedule = (top.promoted is not None
                                   and top.promoted > current_time)

            # If item is promoted and expiration time somewhere in future
            # just reschedule it
            if need_reschedule:
                top.expiry_date = top.promoted
                top.promoted = None
                self._queue.push(top)
            else:
                del self._items[top.session_id]



""" Various static handlers required for SockJS to function properly. """

import hashlib
import random
import sys

from twisted.internet import reactor

from cyclone.web import asynchronous

from sockjs.cyclone.basehandler import BaseHandler, PreflightHandler
from sockjs.cyclone.proto import json_encode


IFRAME_TEXT = '''<!DOCTYPE html>
<html>
<head>
  <meta http-equiv="X-UA-Compatible" content="IE=edge" />
  <meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
  <script>
    document.domain = document.domain;
    _sockjs_onload = function(){SockJS.bootstrap_iframe();};
  </script>
  <script src="%s"></script>
</head>
<body>
  <h2>Don't panic!</h2>
  <p>This is a SockJS hidden iframe. It's used for cross domain magic.</p>
</body>
</html>'''.strip()


class IFrameHandler(BaseHandler):
    """ SockJS IFrame page handler """

    def get(self):
        data = IFRAME_TEXT % self.server.settings['sockjs_url']

        hsh = hashlib.md5(data).hexdigest()

        value = self.request.headers.get('If-None-Match')
        if value and value.find(hsh) != -1:
            # TODO: Fix me? Right now it is a hack to remove content-type
            # header
            self.clear()
            del self._headers['Content-Type']

            self.set_status(304)
            return

        self.enable_cache()

        self.set_header('Etag', hsh)
        self.write(data)


class GreetingsHandler(BaseHandler):
    """ SockJS greetings page handler """

    def initialize(self, server):
        self.server = server

    def get(self):
        self.enable_cache()

        self.set_header('Content-Type', 'text/plain; charset=UTF-8')
        self.write('Welcome to SockJS!\n')


class ChunkingTestHandler(PreflightHandler):
    """ SockJS chunking test handler """

    # Step timeouts according to sockjs documentation
    steps = [0.005, 0.025, 0.125, 0.625, 3.125]

    def initialize(self, server):
        self.server = server
        self.step = 0

    @asynchronous
    def post(self):
        self.preflight()
        self.set_header('Content-Type', 'application/javascript; charset=UTF-8')

        # Send one 'h' immediately
        self.write('h\n')
        self.flush()

        # Send 2048 spaces followed by 'h'
        self.write('%sh\n' % ' ' * 2048)
        self.flush()

        # Send 'h' with different timeouts
        def run_step():
            self.write('h\n')
            self.flush()

            self.step += 1
            if self.step < len(self.steps):
                reactor.callLater(self.steps[self.step], run_step)
            else:
                self.finish()

        reactor.callLater(self.steps[self.step], run_step)


class InfoHandler(PreflightHandler):
    """ SockJS 0.2+ /info handler """
    def initialize(self, server):
        self.server = server
        self.access_methods = 'OPTIONS, GET'

    def get(self):
        self.preflight()
        self.disable_cache()
        self.set_header('Content-Type', 'application/json; charset=UTF-8')

        options = dict(websocket=self.server.websockets_enabled,
                       cookie_needed=self.server.cookie_needed,
                       origins=['*:*'],
                       entropy=random.randint(0, sys.maxint))

        self.write(json_encode(options))



from cyclone.web import asynchronous

from sockjs.cyclone.transports import streamingbase


class XhrStreamingTransport(streamingbase.StreamingTransportBase):
    name = 'xhr_streaming'

    @asynchronous
    def post(self, session_id):
        # Handle cookie
        self.preflight()
        self.handle_session_cookie()
        self.set_header('Content-Type', 'application/javascript; charset=UTF-8')
        self.disable_cache()

        # Send prelude and flush any pending messages
        # prelude is needed to workaround an ie8 weirdness:
        # 'http://blogs.msdn.com/b/ieinternals/archive/2010/04/06/'
        #     'comet-streaming-in-internet-explorer-with-xmlhttprequest-'
        #     'and-xdomainrequest.aspx'
        self.write('h' * 2048 + '\n')
        self.flush()

        if not self._attach_session(session_id):
            self.finish()
            return

        if self.session:
            self.session.flush()

    def connectionLost(self, reason):
        self.session.delayed_close()
        self._detach()

    def send_pack(self, message):
        self.write(message + '\n')
        self.flush()

        # Close connection based on amount of data transferred
        if self.should_finish(len(message) + 1):
            self._detach()
            self.safe_finish()
